extern crate nihav_core;
extern crate nihav_codec_support;
extern crate nihav_registry;
extern crate nihav_allstuff;

use std::fs::File;
use std::io::BufReader;
use nihav_core::io::byteio::{FileReader, ByteReader};
use nihav_core::frame::*;
use nihav_core::options::*;
use nihav_core::codecs::*;
use nihav_core::demuxers::*;
use nihav_core::muxers::*;
use nihav_core::reorder::*;
use nihav_core::scale::*;
use nihav_core::soundcvt::*;
use nihav_registry::detect;
use nihav_registry::register;
use std::env;

mod demux;
use crate::demux::*;
mod null;

fn print_options(name: &str, options: &[NAOptionDefinition]) {
    if options.is_empty() {
        println!("No custom options.");
    } else {
        println!("Options for '{}'", name);
        for opt in options.iter() {
            println!(" {}", opt);
        }
    }
}

struct OptionArgs {
    name:   String,
    value:  Option<String>,
}

struct InputStreamOptions {
    id:             u32,
    drop:           bool,
    dec_opts:       Vec<OptionArgs>,
}

struct OutputStreamOptions {
    id:             u32,
    enc_params:     EncodeParameters,
    enc_name:       String,
    enc_opts:       Vec<OptionArgs>,
}

enum OutputConvert {
    Video(NAScale, NABufferType),
    Audio(NAAudioInfo, NAChannelMap),
    None,
}

#[allow(clippy::large_enum_variant)]
enum OutputMode {
    Drop,
    Copy(u32),
    Encode(u32, Box<dyn NAEncoder>, OutputConvert),
}

#[derive(Default)]
#[allow(clippy::type_complexity)]
struct Transcoder {
    input_name:     String,
    input_fmt:      Option<String>,
    output_name:    String,
    output_fmt:     Option<String>,
    demux_opts:     Vec<OptionArgs>,
    mux_opts:       Vec<OptionArgs>,
    istr_opts:      Vec<InputStreamOptions>,
    ostr_opts:      Vec<OutputStreamOptions>,
    scale_opts:     Vec<(String, String)>,
    decoders:       Vec<Option<(Box<NADecoderSupport>, Box<dyn NADecoder>, Box<dyn FrameReorderer>)>>,
    encoders:       Vec<OutputMode>,
    no_video:       bool,
    no_audio:       bool,
    start:          NATimePoint,
    end:            NATimePoint,
}

macro_rules! parse_and_apply_options {
    ($obj: expr, $in_opts: expr, $name: expr) => {
        let mut opts = Vec::with_capacity($in_opts.len());
        let opt_def = $obj.get_supported_options();
        for opt in $in_opts.iter() {
            let mut found = false;
            for opt_def in opt_def.iter() {
                let mut matches = opt.name == opt_def.name;
                if !matches && opt.name.starts_with("no") {
                    let (_, name) = opt.name.split_at(2);
                    matches = name == opt_def.name;
                }
                if matches {
                    let arg = if let Some(ref str) = opt.value { Some(str) } else { None };
                    let ret = opt_def.parse(&opt.name, arg);
                    if let Ok((val, _)) = ret {
                        opts.push(val);
                    } else {
                        println!("invalid option {} for {}", opt.name, $name);
                    }
                    found = true;
                }
            }
            if !found {
                println!("  ignoring option '{}' for {}", opt.name, $name);
            }
        }
        $obj.set_options(opts.as_slice());
    }
}

impl Transcoder {
    fn new() -> Self { Self::default() }
    fn parse_istream_options(&mut self, opt0: &str, opt1: &str) -> bool {
        let (_, strno) = opt0.split_at(9);
        let ret = strno.parse::<u32>();
        if ret.is_err() { return false; }
        let streamno = ret.unwrap();

        let sidx = if let Some(idx) = self.istr_opts.iter().position(|str| str.id == streamno) {
                idx
            } else {
                self.istr_opts.push(InputStreamOptions {id: streamno, drop: false, dec_opts: Vec::new() });
                self.istr_opts.len() - 1
            };
        let istr = &mut self.istr_opts[sidx];

        for opt in opt1.split(',') {
            let oval: Vec<_> = opt.split('=').collect();
            if oval.len() == 1 {
                match oval[0] {
                    "drop" => { istr.drop = true; },
                    _ => {
                        istr.dec_opts.push(OptionArgs{ name: oval[0].to_string(), value: None });
                    },
                };
            } else if oval.len() == 2 {
                istr.dec_opts.push(OptionArgs{ name: oval[0].to_string(), value: Some(oval[1].to_string()) });
            } else {
                println!("unrecognized option '{}'", opt);
            }
        }
        true
    }
    fn parse_ostream_options(&mut self, opt0: &str, opt1: &str, enc_reg: &RegisteredEncoders) -> bool {
        let (_, strno) = opt0.split_at(9);
        let ret = strno.parse::<u32>();
        if ret.is_err() { return false; }
        let streamno = ret.unwrap();

        let sidx = if let Some(idx) = self.ostr_opts.iter().position(|str| str.id == streamno) {
                idx
            } else {
                self.ostr_opts.push(OutputStreamOptions {id: streamno, enc_name: String::new(), enc_params: EncodeParameters::default(), enc_opts: Vec::new() });
                self.ostr_opts.len() - 1
            };
        let ostr = &mut self.ostr_opts[sidx];

        for opt in opt1.split(',') {
            let oval: Vec<_> = opt.split('=').collect();
            if oval.len() == 1 {
                match oval[0] {
                    "flip" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Video(NAVideoInfo::new(0, 0, false, YUV420_FORMAT));
                        }
                        if let NACodecTypeInfo::Video(ref mut vinfo) = ostr.enc_params.format {
                            vinfo.flipped = true;
                        } else {
                            println!("video option for audio stream");
                        }
                    },
                    "noflip" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Video(NAVideoInfo::new(0, 0, false, YUV420_FORMAT));
                        }
                        if let NACodecTypeInfo::Video(ref mut vinfo) = ostr.enc_params.format {
                            vinfo.flipped = false;
                        } else {
                            println!("video option for audio stream");
                        }
                    },
                    _ => {
                        ostr.enc_opts.push(OptionArgs{ name: oval[0].to_string(), value: None });
                    },
                };
            } else if oval.len() == 2 {
//todo parse encoder options, store, init later
                match oval[0] {
                    "timebase" => {
                        let mut parts = oval[1].split('/');
                        let num = parts.next().unwrap();
                        let den = parts.next();
                        if let Some(den) = den {
                            let rnum = num.parse::<u32>();
                            let rden = den.parse::<u32>();
                            if let (Ok(num), Ok(den)) = (rnum, rden) {
                                ostr.enc_params.tb_num = num;
                                ostr.enc_params.tb_den = den;
                            } else {
                                println!("invalid timebase value");
                            }
                        } else {
                            println!("invalid timebase format (should be num/den)");
                        }
                    },
                    "encoder" => {
                        if enc_reg.find_encoder(oval[1]).is_some() {
                            ostr.enc_name = oval[1].to_string();
                        } else {
                            println!("unknown encoder '{}'", oval[1]);
                            ostr.enc_name = oval[1].to_string();
                        }
                    },
                    "width" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Video(NAVideoInfo::new(0, 0, false, YUV420_FORMAT));
                        }
                        if let NACodecTypeInfo::Video(ref mut vinfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<usize>();
                            if let Ok(val) = ret {
                                vinfo.width = val;
                            } else {
                                println!("invalid width");
                            }
                        } else {
                            println!("video option for audio stream");
                        }
                    },
                    "height" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Video(NAVideoInfo::new(0, 0, false, YUV420_FORMAT));
                        }
                        if let NACodecTypeInfo::Video(ref mut vinfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<usize>();
                            if let Ok(val) = ret {
                                vinfo.height = val;
                            } else {
                                println!("invalid height");
                            }
                        } else {
                            println!("video option for audio stream");
                        }
                    },
                    "pixfmt" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Video(NAVideoInfo::new(0, 0, false, YUV420_FORMAT));
                        }
                        if let NACodecTypeInfo::Video(ref mut vinfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<NAPixelFormaton>();
                            if let Ok(val) = ret {
                                vinfo.format = val;
                            } else {
                                println!("invalid pixel format");
                            }
                        } else {
                            println!("video option for audio stream");
                        }
                    },
                    "srate" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Audio(NAAudioInfo::new(0, 0, SND_S16_FORMAT, 0));
                        }
                        if let NACodecTypeInfo::Audio(ref mut ainfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<u32>();
                            if let Ok(val) = ret {
                                ainfo.sample_rate = val;
                            } else {
                                println!("invalid sampling rate");
                            }
                        } else {
                            println!("audio option for video stream");
                        }
                    },
                    "channels" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Audio(NAAudioInfo::new(0, 0, SND_S16_FORMAT, 0));
                        }
                        if let NACodecTypeInfo::Audio(ref mut ainfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<u8>();
                            if let Ok(val) = ret {
                                ainfo.channels = val;
                            } else {
                                println!("invalid number of channels");
                            }
                        } else {
                            println!("audio option for video stream");
                        }
                    },
                    "block_len" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Audio(NAAudioInfo::new(0, 0, SND_S16_FORMAT, 0));
                        }
                        if let NACodecTypeInfo::Audio(ref mut ainfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<usize>();
                            if let Ok(val) = ret {
                                ainfo.block_len = val;
                            } else {
                                println!("invalid block_length");
                            }
                        } else {
                            println!("audio option for video stream");
                        }
                    },
                    "sfmt" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Audio(NAAudioInfo::new(0, 0, SND_S16_FORMAT, 0));
                        }
                        if let NACodecTypeInfo::Audio(ref mut ainfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<NASoniton>();
                            if let Ok(val) = ret {
                                ainfo.format = val;
                            } else {
                                println!("invalid audio format");
                            }
                        } else {
                            println!("audio option for video stream");
                        }
                    },
// todo channel map negotiation
                    /*"chmap" => {
                        if ostr.enc_params.format == NACodecTypeInfo::None {
                            ostr.enc_params.format = NACodecTypeInfo::Audio(NAAudioInfo::new(0, 0, SND_S16_FORMAT, 0));
                        }
                        if let NACodecTypeInfo::Audio(ref mut ainfo) = ostr.enc_params.format {
                            let ret = oval[1].parse::<NAChannelMap>();
                            if let Ok(val) = ret {
                                ainfo.chmap = val;
                            } else {
                                println!("invalid channel map");
                            }
                        } else {
                            println!("audio option for video stream");
                        }
                    },*/
                    "bitrate" => {
                        let ret = oval[1].parse::<u32>();
                        if let Ok(val) = ret {
                            ostr.enc_params.bitrate = val;
                        } else {
                            println!("invalid bitrate value");
                        }
                    },
                    "quality" => {
                        let ret = oval[1].parse::<u8>();
                        if let Ok(val) = ret {
                            ostr.enc_params.quality = val;
                        } else {
                            println!("invalid quality value");
                        }
                    },
                    _ => {
                        ostr.enc_opts.push(OptionArgs{ name: oval[0].to_string(), value: Some(oval[1].to_string()) });
                    },
                }
            } else {
                println!("unrecognized option '{}'", opt);
            }
        }
        true
    }
    fn parse_demuxer_options(&mut self, opts: &str, dmx_reg: &RegisteredDemuxers) -> bool {
        for opt in opts.split(',') {
            let oval: Vec<_> = opt.split('=').collect();
            if oval.len() == 1 {
                self.demux_opts.push(OptionArgs{ name: oval[0].to_string(), value: None });
            } else if oval.len() == 2 {
                if oval[0] == "format" {
                    if dmx_reg.find_demuxer(oval[1]).is_some() {
                        self.input_fmt = Some(oval[1].to_string());
                    } else {
                        println!("unknown demuxer format '{}'", oval[1]);
                    }
                } else {
                    self.demux_opts.push(OptionArgs{ name: oval[0].to_string(), value: Some(oval[1].to_string()) });
                }
            } else {
                println!("unrecognized option '{}'", opt);
            }
        }
        true
    }
    fn parse_muxer_options(&mut self, opts: &str, mux_reg: &RegisteredMuxers) -> bool {
        for opt in opts.split(',') {
            let oval: Vec<_> = opt.split('=').collect();
            if oval.len() == 1 {
                self.mux_opts.push(OptionArgs{ name: oval[0].to_string(), value: None });
            } else if oval.len() == 2 {
                if oval[0] == "format" {
                    if mux_reg.find_muxer(oval[1]).is_some() {
                        self.output_fmt = Some(oval[1].to_string());
                    } else {
                        println!("unknown muxer format '{}'", oval[1]);
                    }
                } else {
                    self.mux_opts.push(OptionArgs{ name: oval[0].to_string(), value: Some(oval[1].to_string()) });
                }
            } else {
                println!("unrecognized option '{}'", opt);
            }
        }
        true
    }
    fn parse_scale_options(&mut self, opts: &str) -> bool {
        for opt in opts.split(',') {
            let oval: Vec<_> = opt.split('=').collect();
            if oval.len() == 1 {
                self.scale_opts.push((oval[0].to_string(), "".to_string()));
            } else if oval.len() == 2 {
                self.scale_opts.push((oval[0].to_string(), oval[1].to_string()));
            } else {
                println!("unrecognized option '{}'", opt);
                return false;
            }
        }
        true
    }
    fn apply_decoder_options(&self, dec: &mut dyn NADecoder, str_id: u32) {
        if let Some(str_idx) = self.istr_opts.iter().position(|str| str.id == str_id) {
            let dec_opts = dec.get_supported_options();
            if dec_opts.is_empty() { return; }
            let name = format!("input stream {}", str_id);
            parse_and_apply_options!(dec, &self.istr_opts[str_idx].dec_opts, name);
        }
    }
    fn register_output_stream(&mut self, cname: &str, istr: NAStreamRef, out_sm: &mut StreamManager, enc_reg: &RegisteredEncoders) -> bool {
        let out_id = out_sm.get_num_streams() as u32;
        if let Some(str_idx) = self.istr_opts.iter().position(|str| str.id == (istr.get_num() as u32)) {
            if self.istr_opts[str_idx].drop {
                self.encoders.push(OutputMode::Drop);
                return true;
            }
        }

        if let Some(str_idx) = self.ostr_opts.iter().position(|str| str.id == out_id) {
            let oopts = &mut self.ostr_opts[str_idx];
            if oopts.enc_name.as_str() == "copy" && (cname == "any" || istr.get_info().get_name() == cname) {
                out_sm.add_stream_ref(istr);
                self.encoders.push(OutputMode::Copy(out_id));
            } else if cname == "any" || oopts.enc_name.as_str() == cname {
                let enc_create = enc_reg.find_encoder(oopts.enc_name.as_str());
                if enc_create.is_none() {
                    println!("encoder '{}' not found", oopts.enc_name.as_str());
                    return false;
                }
                let mut encoder = (enc_create.unwrap())();
                let forced_out = oopts.enc_params.format != NACodecTypeInfo::None;
                if oopts.enc_params.format == NACodecTypeInfo::None {
                    oopts.enc_params.format = istr.get_info().get_properties();
                }
                if oopts.enc_params.tb_num == 0 {
                    oopts.enc_params.tb_num = istr.tb_num;
                    oopts.enc_params.tb_den = istr.tb_den;
                }
                let ret_eparams = encoder.negotiate_format(&oopts.enc_params);
                if ret_eparams.is_err() {
                    println!("cannot negotiate encoding parameters");
                    return false;
                }
                let ret_eparams = ret_eparams.unwrap();

//todo check for params mismatch
                let cvt = match (&oopts.enc_params.format, &ret_eparams.format) {
                        (NACodecTypeInfo::Video(svinfo), NACodecTypeInfo::Video(dvinfo)) => {
                            if svinfo == dvinfo && !forced_out {
                                OutputConvert::None
                            } else {
                                let ofmt = ScaleInfo { fmt: dvinfo.format, width: dvinfo.width, height: dvinfo.height };
                                let ret = NAScale::new_with_options(ofmt, ofmt, &self.scale_opts);
                                if ret.is_err() {
                                    println!("cannot create scaler");
                                    return false;
                                }
                                let scaler = ret.unwrap();
                                let ret = alloc_video_buffer(*dvinfo, 4);
                                if ret.is_err() {
                                    println!("cannot create scaler buffer");
                                    return false;
                                }
                                let cvt_buf = ret.unwrap();
                                OutputConvert::Video(scaler, cvt_buf)
                            }
                        },
                        (NACodecTypeInfo::Audio(sainfo), NACodecTypeInfo::Audio(dainfo)) => {
                            if sainfo == dainfo {
                                OutputConvert::None
                            } else {
                                let dchmap = match dainfo.channels {
                                        1 => NAChannelMap::from_ms_mapping(0x4),
                                        2 => NAChannelMap::from_ms_mapping(0x3),
                                        _ => {
println!("can't generate default channel map for {} channels", dainfo.channels);
                                            return false;
                                        },
                                    };
//todo channelmap
                                OutputConvert::Audio(*dainfo, dchmap)
                            }
                        },
                        _ => OutputConvert::None,
                    };
                let ret = encoder.init(out_id, ret_eparams);
                if ret.is_err() {
                    println!("error initialising encoder");
                    return false;
                }
                out_sm.add_stream_ref(ret.unwrap());

                let name = format!("output stream {}", out_id);
                parse_and_apply_options!(encoder, &oopts.enc_opts, name);

                self.encoders.push(OutputMode::Encode(out_id, encoder, cvt));
            } else {
println!("encoder {} is not supported by output (expected {})", istr.id, istr.get_info().get_name());
                return false;
            }
        } else if cname == "any" || istr.get_info().get_name() == cname {
            out_sm.add_stream_ref(istr);
            self.encoders.push(OutputMode::Copy(out_id));
        } else {
            let mut oopts = OutputStreamOptions {id: out_id, enc_name: cname.to_owned(), enc_params: EncodeParameters::default(), enc_opts: Vec::new() };

            let enc_create = enc_reg.find_encoder(cname);
            if enc_create.is_none() {
                println!("encoder '{}' not found", oopts.enc_name.as_str());
                return false;
            }
            let mut encoder = (enc_create.unwrap())();
            oopts.enc_params.format = istr.get_info().get_properties();
            oopts.enc_params.tb_num = istr.tb_num;
            oopts.enc_params.tb_den = istr.tb_den;
            let ret_eparams = encoder.negotiate_format(&oopts.enc_params);
            if ret_eparams.is_err() {
                println!("cannot negotiate encoding parameters");
                return false;
            }
            let ret_eparams = ret_eparams.unwrap();

//todo check for params mismatch
            let cvt = match (&oopts.enc_params.format, &ret_eparams.format) {
                    (NACodecTypeInfo::Video(svinfo), NACodecTypeInfo::Video(dvinfo)) => {
                        if svinfo == dvinfo {
                            OutputConvert::None
                        } else {
                            let ofmt = ScaleInfo { fmt: dvinfo.format, width: dvinfo.width, height: dvinfo.height };
                            let ret = NAScale::new_with_options(ofmt, ofmt, &self.scale_opts);
                            if ret.is_err() {
                                println!("cannot create scaler");
                                return false;
                            }
                            let scaler = ret.unwrap();
                            let ret = alloc_video_buffer(*dvinfo, 4);
                            if ret.is_err() {
                                println!("cannot create scaler buffer");
                                return false;
                            }
                            let cvt_buf = ret.unwrap();
                            OutputConvert::Video(scaler, cvt_buf)
                        }
                    },
                    (NACodecTypeInfo::Audio(sainfo), NACodecTypeInfo::Audio(dainfo)) => {
                        if sainfo == dainfo {
                            OutputConvert::None
                        } else {
                            let dchmap = match dainfo.channels {
                                    1 => NAChannelMap::from_ms_mapping(0x4),
                                    2 => NAChannelMap::from_ms_mapping(0x3),
                                    _ => {
println!("can't generate default channel map for {} channels", dainfo.channels);
                                        return false;
                                    },
                                };
//todo channelmap
                            OutputConvert::Audio(*dainfo, dchmap)
                        }
                    },
                    _ => OutputConvert::None,
                };
            let ret = encoder.init(out_id, ret_eparams);
            if ret.is_err() {
                println!("error initialising encoder");
                return false;
            }
            out_sm.add_stream_ref(ret.unwrap());
            self.encoders.push(OutputMode::Encode(out_id, encoder, cvt));
            self.ostr_opts.push(oopts);
        }
        true
    }
    fn map_single(&mut self, cname: &str, ctype: StreamType, src_sm: &StreamManager, out_sm: &mut StreamManager, enc_reg: &RegisteredEncoders) -> bool {
        let mut found_stream = false;
        for istr in src_sm.iter() {
            if istr.get_media_type() != ctype || found_stream {
                self.encoders.push(OutputMode::Drop);
            } else {
                if !self.register_output_stream(cname, istr, out_sm, enc_reg) {
                    return false;
                }
                found_stream = true;
            }
        }
        found_stream
    }
    fn negotiate_stream_map(&mut self, src_sm: &StreamManager, mux_caps: MuxerCapabilities, out_sm: &mut StreamManager, enc_reg: &RegisteredEncoders) -> bool {
        match mux_caps {
            MuxerCapabilities::SingleVideo(cname) => {
                if self.no_video { return false; }
                self.map_single(cname, StreamType::Video, src_sm, out_sm, enc_reg)
            },
            MuxerCapabilities::SingleAudio(cname) => {
                if self.no_audio { return false; }
                self.map_single(cname, StreamType::Audio, src_sm, out_sm, enc_reg)
            },
            MuxerCapabilities::SingleVideoAndAudio(vname, aname) => {
                let mut found_vid = false;
                let mut found_aud = false;
                for istr in src_sm.iter() {
                    if istr.get_media_type() == StreamType::Video && !found_vid && !self.no_video {
                        if !self.register_output_stream(vname, istr, out_sm, enc_reg) {
                            return false;
                        }
                        found_vid = true;
                    } else if istr.get_media_type() == StreamType::Audio && !found_aud && !self.no_audio {
                        if !self.register_output_stream(aname, istr, out_sm, enc_reg) {
                            return false;
                        }
                        found_aud = true;
                    } else {
                        self.encoders.push(OutputMode::Drop);
                    }
                }
                found_vid | found_aud
            },
            MuxerCapabilities::OnlyVideo => {
                if self.no_video { return false; }

                let mut found_vid = false;
                for istr in src_sm.iter() {
                    if istr.get_media_type() == StreamType::Video && !found_vid {
                        if !self.register_output_stream("any", istr, out_sm, enc_reg) {
                            return false;
                        }
                        found_vid = true;
                    } else {
                        self.encoders.push(OutputMode::Drop);
                    }
                }
                found_vid
            },
            MuxerCapabilities::OnlyAudio => {
                if self.no_audio { return false; }

                let mut found_aud = false;
                for istr in src_sm.iter() {
                    if istr.get_media_type() == StreamType::Audio && !found_aud {
                        if !self.register_output_stream("any", istr, out_sm, enc_reg) {
                            return false;
                        }
                        found_aud = true;
                    } else {
                        self.encoders.push(OutputMode::Drop);
                    }
                }
                found_aud
            },
            MuxerCapabilities::Universal => {
                for istr in src_sm.iter() {
                    if (istr.get_media_type() == StreamType::Video && self.no_video) ||
                        (istr.get_media_type() == StreamType::Audio && self.no_audio) {
                        self.encoders.push(OutputMode::Drop);
                        continue;
                    }
                    if !self.register_output_stream("any", istr, out_sm, enc_reg) {
                        return false;
                    }
                }
                true
            },
        }
    }
}

fn encode_frame(dst_id: u32, encoder: &mut Box<dyn NAEncoder>, cvt: &mut OutputConvert, frm: NAFrameRef, scale_opts: &[(String, String)]) -> bool {
    let buf = frm.get_buffer();
    let cbuf = match cvt {
            OutputConvert::None => buf,
            OutputConvert::Video(ref mut scaler, ref mut dbuf) => {
                let cur_ifmt = get_scale_fmt_from_pic(&buf);
                let last_ifmt = scaler.get_in_fmt();
                if cur_ifmt != last_ifmt {
                    let ofmt = scaler.get_out_fmt();
                    let ret = NAScale::new_with_options(cur_ifmt, ofmt, scale_opts);
                    if ret.is_err() {
                        println!("error re-initialising scaler for {} -> {}", cur_ifmt, ofmt);
                        return false;
                    }
                    *scaler = ret.unwrap();
                }
                let ret = scaler.convert(&buf, dbuf);
                if ret.is_err() {
                    println!("error converting frame for encoding");
                    return false;
                }
                dbuf.clone()
            },
            OutputConvert::Audio(ref dinfo, ref dchmap) => {
                let ret = convert_audio_frame(&buf, dinfo, dchmap);
                if ret.is_err() {
                    println!("error converting audio for stream {}", dst_id);
                    return false;
                }
                ret.unwrap()
            },
        };
    let cfrm = NAFrame::new(frm.get_time_information(), frm.frame_type, frm.key, frm.get_info(), cbuf);
    encoder.encode(&cfrm).unwrap();
    true
}

macro_rules! next_arg {
    ($args: expr, $arg_idx: expr) => {
        if $arg_idx + 1 >= $args.len() {
            println!("codec name is required");
        }
        $arg_idx += 1;
    }
}

#[allow(clippy::single_match)]
fn main() {
    let args: Vec<_> = env::args().collect();

    if args.len() == 1 {
        println!("usage: nihav-encoder [options] --input inputfile --output outputfile");
        println!(" use nihav-encoder --help to list all available options");
        return;
    }
    if args.len() == 2 && (args[1] == "--help" || args[1] == "-h") {
        println!("usage: nihav-encoder [options] --input inputfile --output outputfile");
        println!(" query options:");
        println!("  --list-{{decoders,encoders,demuxers,muxers}} - lists all available decoders/encoders/demuxers/muxers");
        println!("  --query-{{decoder,encoder,demuxer,muxer}}-options name - lists all options recognized by that decoder/encoder/demuxer/muxer");
        println!(" processing options:");
        println!("  --input inputfile           - set input file");
        println!("  --input-format fmt          - force input format");
        println!("  --demuxer-options options   - set input demuxer options");
        println!("  --scale-options options     - set scaler options");
        println!("  --output outputfile         - set output file");
        println!("  --output-format fmt         - force output format");
        println!("  --muxer-options options     - set output muxer options");
        println!("  --no-audio                  - do not decode audio streams");
        println!("  --no-video                  - do not decode video streams");
        println!("  --start starttime           - start decoding from given position");
        println!("  --end endtime               - end decoding at given position");
        println!("  --istreamX options          - set options for input stream X");
        println!("  --ostreamX options          - set options for output stream X");
        println!();
        println!(" (de)muxer and stream options are passed as comma-separated list e.g. --ostream0 width=320,height=240,flip");
        return;
    }

    let full_reg = FullRegister::new();

    let mut transcoder = Transcoder::new();

    let mut arg_idx = 1;
    let mut printed_info = false;
    while arg_idx < args.len() {
        match args[arg_idx].as_str() {
            "--list-decoders" => {
                if full_reg.dec_reg.iter().len() > 0 {
                    println!("Registered decoders:");
                    for dec in full_reg.dec_reg.iter() {
                        let cdesc = register::get_codec_description(dec.name);
                        let full_name = if let Some(cd) = cdesc { cd.get_full_name() } else { "???" };
                        println!(" {} ({})", dec.name, full_name);
                    }
                } else {
                    println!("No registered decoders.");
                }
                printed_info = true;
            },
            "--list-encoders" => {
                if full_reg.enc_reg.iter().len() > 0 {
                    println!("Registered encoders:");
                    for enc in full_reg.enc_reg.iter() {
                        let cdesc = register::get_codec_description(enc.name);
                        let full_name = if let Some(cd) = cdesc { cd.get_full_name() } else { "???" };
                        println!(" {} ({})", enc.name, full_name);
                    }
                } else {
                    println!("No registered encoders.");
                }
                printed_info = true;
            },
            "--list-demuxers" => {
                print!("Registered demuxers:");
                for dmx in full_reg.dmx_reg.iter() {
                    print!(" {}", dmx.get_name());
                }
                println!();
                printed_info = true;
            },
            "--list-muxers" => {
                print!("Registered muxers:");
                for mux in full_reg.mux_reg.iter() {
                    print!(" {}", mux.get_name());
                }
                println!();
                printed_info = true;
            },
            "--query-decoder-options" => {
                next_arg!(args, arg_idx);
                let cname = args[arg_idx].as_str();
                if let Some(decfunc) = full_reg.dec_reg.find_decoder(cname) {
                    let dec = (decfunc)();
                    let opts = dec.get_supported_options();
                    print_options(cname, opts);
                } else {
                    println!("codec {} is not found", cname);
                }
                printed_info = true;
            },
            "--query-demuxer-options" => {
                next_arg!(args, arg_idx);
                let dname = args[arg_idx].as_str();
                let mut mr = MemoryReader::new_read(&[]);
                let mut br = ByteReader::new(&mut mr);
                if let Some(dmx_creator) = full_reg.dmx_reg.find_demuxer(dname) {
                    let dmx = dmx_creator.new_demuxer(&mut br);
                    let opts = dmx.get_supported_options();
                    print_options(dname, opts);
                } else {
                    println!("demuxer {} is not found", dname);
                }
                printed_info = true;
            },
            "--query-encoder-options" => {
                next_arg!(args, arg_idx);
                let cname = args[arg_idx].as_str();
                if let Some(encfunc) = full_reg.enc_reg.find_encoder(cname) {
                    let enc = (encfunc)();
                    let opts = enc.get_supported_options();
                    print_options(cname, opts);
                } else {
                    println!("codec {} is not found", cname);
                }
                printed_info = true;
            },
            "--query-muxer-options" => {
                next_arg!(args, arg_idx);
                let name = args[arg_idx].as_str();
                let mut data = [];
                let mut mw = MemoryWriter::new_write(&mut data);
                let mut bw = ByteWriter::new(&mut mw);
                if let Some(mux_creator) = full_reg.mux_reg.find_muxer(name) {
                    let mux = mux_creator.new_muxer(&mut bw);
                    let opts = mux.get_supported_options();
                    print_options(name, opts);
                } else {
                    println!("muxer {} is not found", name);
                }
                printed_info = true;
            },
            "--input" | "-i" => {
                next_arg!(args, arg_idx);
                transcoder.input_name = args[arg_idx].clone();
            },
            "--input-format" => {
                next_arg!(args, arg_idx);
                transcoder.input_fmt = Some(args[arg_idx].clone());
            },
            "--output" | "-o" => {
                next_arg!(args, arg_idx);
                transcoder.output_name = args[arg_idx].clone();
            },
            "--output-format" => {
                next_arg!(args, arg_idx);
                transcoder.output_fmt = Some(args[arg_idx].clone());
            },
            "--demuxer-options" => {
                next_arg!(args, arg_idx);
                if !transcoder.parse_demuxer_options(&args[arg_idx], &full_reg.dmx_reg) {
                    println!("invalid demuxer option syntax");
                    return;
                }
            },
            "--scale-options" => {
                next_arg!(args, arg_idx);
                if !transcoder.parse_scale_options(&args[arg_idx]) {
                    println!("invalid scale option syntax");
                    return;
                }
            },
            "--no-video" | "-vn" => {
                transcoder.no_video = true;
            },
            "--no-audio" | "-an" => {
                transcoder.no_audio = true;
            },
            "--start" => {
                next_arg!(args, arg_idx);
                let ret = args[arg_idx].parse::<NATimePoint>();
                if let Ok(val) = ret {
                    transcoder.start = val;
                } else {
                    println!("invalid start time");
                    return;
                }
            },
            "--end" => {
                next_arg!(args, arg_idx);
                let ret = args[arg_idx].parse::<NATimePoint>();
                if let Ok(val) = ret {
                    transcoder.end = val;
                } else {
                    println!("invalid end time");
                    return;
                }
            },
            "--muxer-options" => {
                next_arg!(args, arg_idx);
                if !transcoder.parse_muxer_options(&args[arg_idx], &full_reg.mux_reg) {
                    println!("invalid muxer option syntax");
                    return;
                }
            },
            _ => {
                if args[arg_idx].starts_with("--istream") {
                    let opt0 = &args[arg_idx];
                    next_arg!(args, arg_idx);
                    if !transcoder.parse_istream_options(opt0, &args[arg_idx]) {
                        println!("invalid input stream option syntax");
                        return;
                    }
                } else if args[arg_idx].starts_with("--ostream") {
                    let opt0 = &args[arg_idx];
                    next_arg!(args, arg_idx);
                    if !transcoder.parse_ostream_options(opt0, &args[arg_idx], &full_reg.enc_reg) {
                        println!("invalid output stream option syntax");
                        return;
                    }
                } else if args[arg_idx].starts_with("--") {
                    println!("unknown option");
                } else {
                    println!("unrecognized argument");
                }
            },
        };
        arg_idx += 1;
    }

    if printed_info {
        return;
    }

    if transcoder.input_name.is_empty() {
        println!("no input name provided");
        return;
    }
    if transcoder.output_name.is_empty() {
        println!("no output name provided");
        return;
    }

    let res = File::open(transcoder.input_name.as_str());
    if res.is_err() {
        println!("error opening input");
        return;
    }
    let file = res.unwrap();
    let file = BufReader::new(file);
    let mut fr = FileReader::new_read(file);
    let mut br = ByteReader::new(&mut fr);
    let (is_raw, start, end) = if transcoder.input_fmt.is_none() {
            detect_tags(&mut br)
        } else {
            (false, 0, None)
        };

    let mut nfr: Box<dyn ByteIO>;
    if start != 0 || end.is_some() {
        let file = fr.finish();
        nfr = Box::new(BoundedFileReader::new_read(file, start, end).unwrap());
    } else {
        nfr = Box::new(fr);
    }
    let mut br = ByteReader::new(nfr.as_mut());

    let mut dmx = DemuxerObject::create(&mut br, &full_reg, transcoder.input_name.as_str(), &transcoder.input_fmt, is_raw);
    if dmx.is_none() {
        println!("cannot find demuxer for '{}'", transcoder.input_name.as_str());
        return;
    }
    parse_and_apply_options!(dmx, &transcoder.demux_opts, "input");
    for i in 0..dmx.get_num_streams() {
        let s = dmx.get_stream(i).unwrap();
        let info = s.get_info();
        let decfunc = full_reg.dec_reg.find_decoder(info.get_name());
println!("stream {} - {} {}", i, s, info.get_name());
        let str_id = s.get_num() as u32;
        if let Some(create_dec) = decfunc {
            let mut dec = (create_dec)();
            let mut dsupp = Box::new(NADecoderSupport::new());
            let ret = dec.init(&mut dsupp, info.clone());
            if ret.is_err() {
                println!("Error initialising decoder '{}' for stream {}", info.get_name(), str_id);
                return;
            }
            transcoder.apply_decoder_options(dec.as_mut(), str_id);
            let desc = register::get_codec_description(info.get_name());
            let has_b = if let Some(desc) = desc {
                    desc.has_reorder()
                } else {
                    println!("No codec description found, using B-frame reorderer.");
                    true
                };
            let reord: Box<dyn FrameReorderer> = if has_b { Box::new(IPBReorderer::new()) } else { Box::new(NoReorderer::new()) };
            transcoder.decoders.push(Some((dsupp, dec, reord)));
        } else {
            println!("No decoder for stream {} ({}) is found", str_id, info.get_name());
            transcoder.decoders.push(None);
        }
    }
    if transcoder.start != NATimePoint::None {
        let ret = dmx.seek(transcoder.start);
        if ret.is_err() {
            println!(" failed to seek to {} error {:?}", transcoder.start, ret.err().unwrap());
        }
    }

    let output_fmt = if let Some(ref str) = transcoder.output_fmt {
            str
        } else if let Some(str) = detect::detect_format_by_name(transcoder.output_name.as_str()) {
            str
        } else {
            println!("Cannot guess muxer for output");
            return;
        };
    let ret = full_reg.mux_reg.find_muxer(output_fmt);
    let ofmt = output_fmt.to_string();

    if ret.is_none() {
        println!("cannot find muxer '{}'", output_fmt);
    }
    let mux_creator = ret.unwrap();

    let mux_caps = mux_creator.get_capabilities();
    let mut out_sm = StreamManager::new();
    if !transcoder.negotiate_stream_map(dmx.get_stream_manager(), mux_caps, &mut out_sm, &full_reg.enc_reg) {
        println!("cannot determine stream map");
        return;
    }

    let ret = File::create(transcoder.output_name.as_str());
    if ret.is_err() {
        println!("cannot open output file");
        return;
    }
    let mut fw = FileWriter::new_write(ret.unwrap());
    let mut bw = ByteWriter::new(&mut fw);
    let ret = create_muxer(mux_creator, out_sm, &mut bw);
    if let Err(err) = ret {
        println!("cannot create muxer instance {:?}", err);
        return;
    }
    let mut mux = ret.unwrap();
    parse_and_apply_options!(mux, &transcoder.mux_opts, "output");

    println!("Output {} muxer {}", transcoder.output_name, ofmt);
    for ostr in mux.get_streams() {
        println!(" #{}: {} {}", ostr.get_num(), ostr, ostr.get_info().get_name());
    }

    'main_loop: loop {
        let pktres = dmx.get_frame();
        if let Err(DemuxerError::EOF) = pktres { break; }
        if pktres.is_err() {
            println!("demuxing error");
            break;
        }
        let mut pkt = pktres.unwrap();
        if transcoder.start != NATimePoint::None && pkt.ts.less_than(transcoder.start) { continue; }
        let src_id = pkt.get_stream().get_num();
        match transcoder.encoders[src_id] {
            OutputMode::Drop => {},
            OutputMode::Copy(dst_id) => {
                let dstr = mux.get_stream(dst_id as usize).unwrap();
                pkt.reassign(dstr, pkt.get_time_information());
                if transcoder.end != NATimePoint::None && !pkt.ts.less_than(transcoder.end) { break 'main_loop; }
                if mux.mux_frame(pkt).is_err() {
                    println!("error muxing packet");
                    break;
                }
            },
            OutputMode::Encode(dst_id, ref mut encoder, ref mut cvt) => {
                if let Some((ref mut dsupp, ref mut decoder, ref mut reorderer)) = transcoder.decoders[src_id] {
                    let ret = decoder.decode(dsupp, &pkt);
                    if let (true, Err(DecoderError::MissingReference)) = (transcoder.start != NATimePoint::None, &ret) {
                        continue;
                    }
                    if ret.is_err() {
                        println!("error decoding stream {}", src_id);
                        break;
                    }
                    let frm = ret.unwrap();
                    reorderer.add_frame(frm);
                    while let Some(frm) = reorderer.get_frame() {
                        if !encode_frame(dst_id, encoder, cvt, frm, &transcoder.scale_opts) {
                            break;
                        }
                        while let Ok(Some(pkt)) = encoder.get_packet() {
                            if transcoder.end != NATimePoint::None && !pkt.ts.less_than(transcoder.end) { break 'main_loop; }
                            mux.mux_frame(pkt).unwrap();
                        }
                    }
                } else {
                    println!("no decoder for stream {}", src_id);
                    break;
                }
            },
        };
    }
    'reord_flush_loop: for str in dmx.get_streams() {
        let src_id = str.get_num();
        if let OutputMode::Encode(dst_id, ref mut encoder, ref mut cvt) = transcoder.encoders[src_id] {
            if let Some((_, _, ref mut reorderer)) = transcoder.decoders[src_id] {
                while let Some(frm) = reorderer.get_last_frames() {
                    if !encode_frame(dst_id, encoder, cvt, frm, &transcoder.scale_opts) {
                        break;
                    }
                    while let Ok(Some(pkt)) = encoder.get_packet() {
                        if transcoder.end != NATimePoint::None && !pkt.ts.less_than(transcoder.end) { break 'reord_flush_loop; }
                        mux.mux_frame(pkt).unwrap();
                    }
                }
            }
        }
    }
    'flush_loop: for enc in transcoder.encoders.iter_mut() {
        match enc {
            OutputMode::Encode(str_id, ref mut encoder, _) => {
                let ret = encoder.flush();
                if ret.is_err() {
                    println!("error flushing encoder for stream {}", str_id);
                    break;
                } else {
                    while let Ok(Some(pkt)) = encoder.get_packet() {
                        if mux.mux_frame(pkt).is_err() {
                            println!("error muxing packet");
                            break 'flush_loop;
                        }
                    }
                }
            },
            _ => {},
        };
    }

    let ret = mux.end();
    if ret.is_err() {
        println!("error at finalising muxing");
    }
}
